//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Axis/FixedBinAxis.cpp
//! @brief     Implement class FixedBinAxis.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Base/Axis/FixedBinAxis.h"
#include "Base/Util/Algorithms.h"
#include "Base/Util/Assert.h"
#include <iomanip>
#include <limits>

namespace {

std::vector<Bin1D> bounds2bins(size_t nbins, double start, double end)
{
    ASSERT(nbins > 0);
    ASSERT(start <= end);
    const double width = (end - start) / nbins;
    std::vector<Bin1D> result;
    for (size_t i = 0; i < nbins; ++i)
        result.emplace_back(Bin1D::FromTo(start + i * width, start + (i + 1) * width));
    ASSERT(result.size() == nbins);
    return result;
}

} // namespace


FixedBinAxis::FixedBinAxis(const std::string& name, size_t nbins, double start, double end)
    : IAxis(name, bounds2bins(nbins, start, end))
    , m_nbins(nbins)
    , m_start(start)
    , m_end(end)
{
}

FixedBinAxis* FixedBinAxis::clone() const
{
    auto* result = new FixedBinAxis(axisName(), m_nbins, m_start, m_end);
    return result;
}

void FixedBinAxis::clip(double lower, double upper)
{
    if (lower >= upper)
        throw std::runtime_error(
            "FixedBinAxis::clip -> Error. 'lower' should be smaller than 'upper'");

    if (lower < min())
        lower = bin(0).center();
    if (upper >= max())
        upper = bin(size() - 1).center();

    const size_t nbin1 = closestIndex(lower);
    const size_t nbin2 = closestIndex(upper);

    // create tmp vars until everything is calculated, otherwise the calculation will be corrupted
    // by partially changed values
    const auto newStart = bin(nbin1).lowerBound();
    const auto newEnd = bin(nbin2).upperBound();

    m_nbins = nbin2 - nbin1 + 1;
    m_start = newStart;
    m_end = newEnd;

    m_bins = bounds2bins(m_nbins, m_start, m_end);
}

void FixedBinAxis::print(std::ostream& ostr) const
{
    ostr << "FixedBinAxis(\"" << axisName() << "\", " << size() << ", "
         << std::setprecision(std::numeric_limits<double>::digits10 + 2) << min() << ", " << max()
         << ")";
}

bool FixedBinAxis::equals(const IAxis& other) const
{
    if (!IAxis::equals(other))
        return false;
    if (const auto* otherAxis = dynamic_cast<const FixedBinAxis*>(&other)) {
        if (size() != otherAxis->size())
            return false;
        if (!BaseUtils::algo::almostEqual(m_start, otherAxis->m_start))
            return false;
        if (!BaseUtils::algo::almostEqual(m_end, otherAxis->m_end))
            return false;
        return true;
    }
    return false;
}

//... In global namespace.

FixedBinAxis FixedScanAxis(const std::string& name, size_t nbins, double scan_start,
                           double scan_end)
{
    if (nbins < 2)
        throw std::runtime_error("Cannot create scan axis with less than two points");
    const double delta = (scan_end - scan_start) / (nbins - 1);
    return FixedBinAxis(name, nbins, scan_start - delta / 2, scan_end + delta / 2);
}
