//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Axis/IAxis.h
//! @brief     Defines interface IAxis.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_BASE_AXIS_IAXIS_H
#define BORNAGAIN_BASE_AXIS_IAXIS_H

#include "Base/Axis/Bin.h"
#include <string>
#include <utility>
#include <vector>

//! Abstract base class for one-dimensional axes.

class IAxis {
public:
    IAxis(std::string name, const std::vector<Bin1D>& bins);
    IAxis(const IAxis&) = delete;
    IAxis(IAxis&&) = default;
    virtual ~IAxis();

    virtual IAxis* clone() const = 0;

    //! Sets the axis label
    void setAxisName(std::string name) { m_name = name; }

    //! Returns the label of the axis
    std::string axisName() const { return m_name; }

    //... Getter functions: range

    //! Returns the number of bins
    size_t size() const;

    //! Returns value of first point of axis
    double min() const;

    //! Returns value of last point of axis
    double max() const;

    //! Returns lower and upper bound in a pair.
    std::pair<double, double> bounds() const;

    //! Returns true if axis contains given point
    bool rangeComprises(double value) const;

    //! Returns distance from first to last point
    double span() const;

    //! Returns midpoint of axis
    double center() const;

    //... Getter functions: bins

    const Bin1D& bin(size_t i) const;
    double binCenter(size_t i) const;

    std::vector<double> binCenters() const;

    //! find bin index which is best match for given value
    size_t closestIndex(double value) const;

    //! Clips this axis to the given values
    virtual void clip(double lower, double upper);

    //! Convenience overload to clip this axis to the given values.
    //! bounds.first is lower, bounds.second is upper value.
    void clip(std::pair<double, double> bounds);

    //! test for equality
    bool operator==(const IAxis& right) const { return equals(right); }
    bool operator!=(const IAxis& right) const { return !(*this == right); }

    friend std::ostream& operator<<(std::ostream& ostr, const IAxis& m)
    {
        m.print(ostr);
        return ostr;
    }

protected:
    virtual void print(std::ostream& ostr) const = 0;
    virtual bool equals(const IAxis& other) const; // overloaded in child classes

    // private:
    std::string m_name;
    std::vector<Bin1D> m_bins;
};

#endif // BORNAGAIN_BASE_AXIS_IAXIS_H
